/*
 * speaker.c - PC Speaker Driver + Emulator
 *
 *  David Olofson, 2001
 */

#include <stdlib.h>
#include "speaker.h"
#include "SDL.h"
#include "SDL_audio.h"
#include "SDL_thread.h"

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/kd.h>
#include <X11/Xlib.h>	
#ifndef CLOCK_TICK_RATE
#define CLOCK_TICK_RATE 1193180
#endif


static void (*tickhandler)(void);
int freq = 0;
int old_freq = -1;
int per = 17;
int do_retrig = 0;
int running = 0;
SDL_Thread *driver_thread;
int SPK_Target = SPK_TARGET_NONE;

/*
 * Speaker emulator for SDL Audio; all targets
 */

static SDL_AudioSpec audiospec;

typedef struct
{
	int	phase;
	int	k;
	int	facc;
} osc_t;

static osc_t osc = {0,0,0};
static int timer = 0;
static int timer_period = 441;

static short *delaybuf;
static int delaypos;
static int fbdelay, rdelay, ldelay;

static int hpf_shift = 5;

#define	_DB(x)	delaybuf[(delaypos - (x)) & 0xffff]

/* Callback. */
static void SPK_Driver_aud(void *ud, Uint8 *stream, int len)
{
	int s = 0;
	int oscout;
	Sint16 *buf = (Sint16 *)stream;
	len >>= 2;
	while(s < len)
	{
		int s_end = s + timer;
		if(s_end > len-1)
			s_end = len-1;
		timer -= s_end - s;
		if(do_retrig)
		{
			osc.k = 0;
			osc.phase = 0;
			do_retrig = 0;
		}
		if(freq)
			osc.k = (freq<<16) / audiospec.freq;
		while(s++ < s_end)
		{
			/* Oscillator */
			osc.phase += osc.k;
			oscout = (osc.phase & 0x8000) - 16384;

			/* IIR HP filter */
			osc.facc += (oscout - osc.facc) >> hpf_shift;
			oscout = oscout - osc.facc;

			/* Mix dry output */
			buf[s<<1] += oscout >> 1;
			buf[(s<<1)+1] += oscout >> 1;

			/* Add stereo feedback delay */
			_DB(0) = (oscout + _DB(fbdelay)) >> 1;
			buf[s<<1] += _DB(ldelay) >> 2;
			buf[(s<<1)+1] += _DB(rdelay) >> 2;
			++delaypos;
		}
		/* Periodic callback */
		if(timer <= 0)
		{
			if(tickhandler)
				tickhandler();
			timer = timer_period;
		}
	}
}


int SPK_Init_aud(int period, int bass)
{
	SDL_AudioSpec as;

	delaybuf = calloc(65536, sizeof(short));
	if(!delaybuf)
		return -1;

	if(SDL_InitSubSystem(SDL_INIT_AUDIO) < 0)
		return -2;

	as.freq = 44100;
	as.format = AUDIO_S16SYS;
	as.channels = 2;
	as.samples = 1024;
	as.callback = SPK_Driver_aud;
	if(SDL_OpenAudio(&as, &audiospec) < 0)
		return -3;

	if(audiospec.format != AUDIO_S16SYS)
		return -4;

	SPK_NoSound();
	timer_period = audiospec.freq * period / 1000;

	if(audiospec.freq < 3500)
		hpf_shift = 1;
	else if(audiospec.freq < 7000)
		hpf_shift = 1;
	else if(audiospec.freq < 15000)
		hpf_shift = 2;
	else if(audiospec.freq < 30000)
		hpf_shift = 3;
	else if(audiospec.freq < 60000)
		hpf_shift = 4;
	else
		hpf_shift = 5;

	hpf_shift += bass;

	fbdelay = audiospec.freq * 0.4;
	rdelay = audiospec.freq * 0.28;
	ldelay = audiospec.freq * 0.32;

	SDL_PauseAudio(0);
	return 0;
}


/*
 * Speaker driver for Linux console; Linux only (?)
 */

int con_fd = 0;

/* Thread. */
static int SPK_Driver_con(void *data)
{
	while(running)
	{
		if(do_retrig)
		{
			ioctl(con_fd, KIOCSOUND, 0);
			SDL_Delay(per/2);
		}
		if(freq)
			ioctl(con_fd, KIOCSOUND,
					CLOCK_TICK_RATE/freq);
		else if(!do_retrig)
			ioctl(con_fd, KIOCSOUND, 0);

		if(do_retrig)
			SDL_Delay(per/2);
		else
			SDL_Delay(per);

		old_freq = freq;
		do_retrig = 0;

		/* Periodic callback */
		if(tickhandler)
			tickhandler();
	}
	return 0;
}

int SPK_Init_con(int period)
{
	per = period;
	con_fd = open("/dev/console", O_WRONLY);
	if(con_fd < 0)
		return -1;
	running = 1;
	driver_thread = SDL_CreateThread(SPK_Driver_con, NULL);
	if(!driver_thread)
		return -2;
	return 0;
}


/*
 * Speaker driver for XBell(); all X targets
 * (although some will not do what we want...)
 */

Display *x_disp;
XKeyboardState x_kbdstate_save;

/* Thread. */
static int SPK_Driver_x(void *data)
{
	while(running)
	{
		if(do_retrig)
		{
			XKeyboardControl ctrl;
			ctrl.bell_pitch = 50;
			ctrl.bell_duration = 0;
			XChangeKeyboardControl(x_disp,
						KBBellDuration |
						KBBellPitch,
						&ctrl);
			XBell(x_disp, 100);
			XFlush(x_disp);
			SDL_Delay(per/2);
		}
		if(freq)
		{
			if(freq != old_freq)
			{
				XKeyboardControl ctrl;
				ctrl.bell_pitch = freq;
				ctrl.bell_duration = per;
				XChangeKeyboardControl(x_disp,
						KBBellDuration |
						KBBellPitch,
						&ctrl);
				old_freq = freq;
			}

			XBell(x_disp, 100);
			XFlush(x_disp);
		}

		if(do_retrig)
			SDL_Delay(per/2);
		else
			SDL_Delay(per);

		do_retrig = 0;

		/* Periodic callback */
		if(tickhandler)
			tickhandler();
	}
	return 0;
}


int SPK_Init_x(int period)
{
	per = period;

	x_disp = XOpenDisplay(NULL);
	if(!x_disp)
		return -1;

	XGetKeyboardControl(x_disp, &x_kbdstate_save);

	running = 1;
	driver_thread = SDL_CreateThread(SPK_Driver_x, NULL);
	if(!driver_thread)
		return -2;
	return 0;
}



/*
 * Common stuff
 */

static int try_init(int target, int period, int bass)
{
	int ret = -1;
	SPK_Target = target;
	switch(SPK_Target)
	{
	  case SPK_TARGET_NONE:
		ret = 0;	/* Can't fail... */
		break;
	  case SPK_TARGET_AUDIO:
		ret = SPK_Init_aud(period, bass);
		break;
	  case SPK_TARGET_CONSOLE:
		ret = SPK_Init_con(period);
		break;
	  case SPK_TARGET_XBELL:
		ret = SPK_Init_x(period);
		break;
	}
	if(ret < 0)
		SPK_Target = SPK_TARGET_NONE;
	return ret;
}

static int try_init_all(int period, int bass, int flags)
{
	if(flags & SPK_PREFER_AUDIO)
		if(try_init(SPK_TARGET_AUDIO, period, bass) == 0)
			return 0;
	if(try_init(SPK_TARGET_CONSOLE, period, bass) == 0)
		return 0;
	if(try_init(SPK_TARGET_XBELL, period, bass) == 0)
		return 0;
	if(!(flags & SPK_PREFER_AUDIO))
		if(try_init(SPK_TARGET_AUDIO, period, bass) == 0)
			return 0;
	return -1;
}

int SPK_Init(int period, int bass, int flags)
{
	int ret = try_init_all(period, bass, flags);
	switch(SPK_Target)
	{
	  case SPK_TARGET_NONE:
		printf("SPK: Failed to initialize speaker sound!\n");
		break;
	  case SPK_TARGET_AUDIO:
		printf("SPK: Using audio speaker emulation.\n");
		break;
	  case SPK_TARGET_CONSOLE:
		printf("SPK: Using console driver.\n");
		break;
	  case SPK_TARGET_XBELL:
		printf("SPK: Using XBell driver.\n");
		break;
	}
	return ret;
}


void SPK_SetTickHandler(void (*handler)(void))
{
	tickhandler = handler;
}


static void stop_driver(void)
{
	int status;
	/*
	 * This doesn't exactly seem bullet proof to me...
	 */
	running = 0;
	SDL_WaitThread(driver_thread, &status);
}


void SPK_Close(void)
{
	XKeyboardControl ctrl;
	switch(SPK_Target)
	{
	  case SPK_TARGET_NONE:
		break;
	  case SPK_TARGET_AUDIO:
		SDL_CloseAudio();
		SDL_QuitSubSystem(SDL_INIT_AUDIO);
		break;
	  case SPK_TARGET_CONSOLE:
		stop_driver();
		ioctl(con_fd, KIOCSOUND, 0);
		close(con_fd);
		break;
	  case SPK_TARGET_XBELL:
		stop_driver();
		ctrl.bell_pitch = freq;
		ctrl.bell_duration = per;
		XChangeKeyboardControl(x_disp,
				KBBellDuration | KBBellPitch,
				&ctrl);
		XCloseDisplay(x_disp);
		break;
	}
}


/*
 * Set sound frequency.
 * (Starts oscillator from phase 0
 * unless it's already playing.)
 */
void SPK_Sound(int frequency)
{
	freq = frequency;
}


void SPK_NoSound(void)
{
	do_retrig = 1;
	freq = 0;
}
